--- Scanf.e -------------------------------------------------------------------
-- A scanf() function for Euphoria.
-- -------------------------------------------------------------------------
-- Description:
-- scanf() is s useful little function that exists in C but not in
-- Euphoria. Broadly, it is used to obtain data that is supplied in the
-- form of a formatted text-string, using a programmer-supplied format
-- string to check compliance with the format and to assist retrieval
-- of data from the text string.

-- Users of C will be familiar with scanf(). Those not so familiar
-- should consult a book on C.

-- scanf() will try to make sense of the format string and data text string
-- you submit to it and return the suitable values. If there is serious mis-
-- match between the format string and the text string, the call will write
-- a diagnostic message on a DOS screen, and execute abort(1) to terminate
-- the program.

-- SEE also the short introductory notes to the scanf() function below.

-- --------------------------------------------------------------------------

include misc.e
include get.e

type string(sequence s)
	for i = 1 to length(s) do
		if not integer(s[i]) then
			return 0	-- fails
		end if
	end for
	return 1			-- success
end type

type non_null_string(sequence s)
	if string(s) and length(s) > 0 then
		return 1	-- success
	else
		return 0	-- fails
	end if
end type

procedure info_abort(sequence msg, sequence datatxt, sequence fmat)
-- Displays a DOS error message and halts program execution
	msg &= sprintf("\n\nFORMAT: \"%s\"\nDATA  : \"%s\"",{fmat,datatxt})
	msg &= "\n\nProgram will close.\n\n"
	printf(1, "%s\n",{msg})
	system("Pause",0)
	abort(1)
end procedure

function scanf_count_params(sequence fmat)
-- Counts number of params in fmat by counting '%' chars
	atom count
	count = 0
	for i=1 to length(fmat) do
		if fmat[i] = '%' then
			count += 1
		end if
	end for
	return count
end function	-- scanf_count_params

function scanf_scan(non_null_string fmat, non_null_string datatxt)
-- Extracts info for scanf() function
-- Entry:	fmat	= format string
--		   datatxt = string carrying data to be extracted
-- Returns:  res	 = sequence containing the extracted data
--			else a failure message
	sequence res,word,paramsymbols
	integer fpoint,datpoint,fchar,datachar,paramflag,wrdend,wrdendsymb,
			paramcnt,fparamcount
	res = {}
	paramcnt = 0
	fmat &= ' '	   -- add guards
	datatxt &= ' '
	fparamcount = scanf_count_params(fmat)

	if fparamcount > 0 then

		-- maybe a correct format string - go ahead
		paramsymbols = "sdfgx"				 -- init outer loop
		datpoint = 1
		while datpoint < length(datatxt) do	-- for each cycle thru fmat
			fpoint = 1						 -- init inner loop
			paramflag = 0

		  while fpoint < length(fmat) do	   -- process thru the fmat string
			fchar = fmat[fpoint]			   -- fetch char from fmat
			-- PATCH
			if datpoint >= length(datatxt) then
				exit
			end if
			datachar = datatxt[datpoint]	   -- fetch char from text
			if fchar = '%' then				-- PARAMETER START?
				-- will get param symbol on next cycle
				paramflag = 1				  -- flag for next cycle
				paramcnt += 1
			elsif paramflag then
				-- complete work on a parameter
				paramflag = 0		   	   -- reset
				-- process a word
				if find(fchar,paramsymbols) then
					-- legal param, go ahead and get the word
					wrdendsymb = fmat[fpoint+1]
					wrdend = find(wrdendsymb,datatxt[datpoint..$])
					if wrdend > 0 and wrdend then
						wrdend += datpoint-2
						if wrdend <= length(datatxt) then
							word = datatxt[datpoint-1..wrdend]

							-- save the word, its type, and where found
							res = append(res,{word, fchar, paramcnt})

							datpoint = wrdend		-- advance in text
						else
							-- error
							info_abort("Scanf error: Scans past end of data string.",datatxt,fmat)

						end if
					else
						-- some other problem
						info_abort("Scanf error: Uncertain scanning problem.",datatxt,fmat)
					end if
				else
					-- unknown param in fmat
					info_abort(sprintf("Scanf error: Unknown param (# %d) in format string.",paramcnt),datatxt,fmat)
				end if
			else
				if fchar != datachar then
					-- we are moving through fmat string, not in a param,
					-- but texts have mismatched.
					info_abort("Scanf error: Format string doesn't match data string.",datatxt,fmat)
				end if

			end if

			fpoint +=1		-- advance for next time around this loop
			datpoint += 1

		  end while
		end while
	else
		-- improper format string; no params in it.
		info_abort("Scanf error: Bad format string, has no params.",datatxt,fmat)
	end if

	if paramcnt < fparamcount then
		-- error - have missed data
		info_abort("Scanf error: Could not read all data in data string.\n" &
				   "Format string doesn't match data string.",datatxt,fmat)
	end if

	return res
end function	-- scanf_scan

function scanf_eval(sequence res, non_null_string fmat, non_null_string datatxt)
-- Extracts the data in res to appropriate forms
	sequence word, v
	integer symbol

	for i = 1 to length(res) do
		-- process word according to param symbol

		word = res[i][1]
		symbol = res[i][2]

		if symbol = 's' then							-- string?
			-- yes, just add to result
			res[i] = word
		elsif symbol = 'd' or symbol = 'x' then		 -- integer or hex?
			if symbol = 'x' and word[1] != '#' then
				-- bad value, not hex
				info_abort(sprintf("Scanf error: Type mismatch at param # %d",res[i][3]),datatxt,fmat)
			end if
			if not find('.',word) then 				 -- not decimal?
				-- good -- get its value
				v = value(word)
				if v[1] = GET_SUCCESS then
					-- good value - add to result
					res[i] = v[2]
				else
					-- bad value
				info_abort(sprintf("Scanf error: Type mismatch or bad value at param # %d",res[i][3]),datatxt,fmat)
				end if
			else
				-- bad value, probably decimal
				info_abort(sprintf("Scanf error: Type mismatch at param # %d",res[i][3]),datatxt,fmat)

			end if
		elsif symbol = 'f' or symbol = 'g' then		 		-- decimal?
			-- ok -- get its value
			v = value(word)
			if v[1] = GET_SUCCESS then
				-- good value - add numeric to result
				res[i] = v[2]
			else
				-- bad value, probably string
				info_abort(sprintf("Scanf error: Type mismatch at param # %d",res[i][3]),datatxt,fmat)
			end if
		end if

	end for

	return res
end function	-- scanf_eval

global function scanf(non_null_string fmat, non_null_string datatxt)
-- Extracts info for scanf() function.
-- Recognizes, for reading purposes, format indicators: %s, %d, %f, %g, %x
-- Entry:	fmat	= format string.
--		   datatxt = string carrying formatted data to be extracted.
-- Returns:  res	 = sequence containing the extracted data.
--		   program aborts(1) with diagnostic message if extraction fails.
-- Notes:
--  Hex (%x) values are returned as decimal integers; scanning expects to
--		   find such numbers preceded by '#' in the datatxt, and will
--		   return an error if this is not found.
-- This function is far from perfect, but works quite well.
	sequence res

	res = scanf_scan(fmat, datatxt)		 -- extract formatted text
	res = scanf_eval(res, fmat, datatxt)	-- evaluate the text we got
	return res
end function	-- scanf

procedure test()
-- These (with modifications) were used in testing
	? scanf("%s//%d//%f","ABCD//-10//1.0e-30")
	--? scanf("%d ", "-10 22 650 650")
	--? scanf("%s|%f|%f", "12|1.234|2.468")
	system("pause",0)
end procedure

test()

